@@ -194,7 +194,7 @@ static void Mpack_Flush(Writer* writer, const char* buffer, size_t bytes) {
 ////////////////////////////////////////////////////////////////////////////////
 // Misc
 ////////////////////////////////////////////////////////////////////////////////
-static Timestamp Fprofile_Time(void) {
+static inline Timestamp Fprofile_Time(void) {
     // This effectively fetches the output of RDTSC, making it quick but not
     // very useful on its own, as there is no meaningful epoch for the time.
     Timestamp tsc = { 0 };
@@ -202,6 +202,24 @@ static Timestamp Fprofile_Time(void) {
     return tsc;
 }
 
+// Given a PyObject (likely `co_name` or similar), return a UTF-8
+// representation.
+static inline const char* Fprofile_UnicodeToUtf8(PyObject* obj) {
+    if (obj == NULL) {
+        return "<NULL>";
+    }
+
+    if (PyUnicode_Check(obj)) {
+        const char* utf8 = PyUnicode_AsUTF8(obj);
+
+        return utf8 != NULL ? utf8 : "<DECODE ERROR>";
+    } else if (obj == Py_None) {
+        return "<NONE>";
+    } else {
+        return "<UNKNOWN>";
+    }
+}
+
 ////////////////////////////////////////////////////////////////////////////////
 // Function Hooking
 ////////////////////////////////////////////////////////////////////////////////
@@ -471,11 +489,27 @@ static int Fprofile_FunctionTrace(PyObject* obj, PyFrameObject* frame, int what,
             mpack_finish_array(writer);
 
             {
-                PyObject* name     = frame->f_code->co_name;
-                PyObject* filename = frame->f_code->co_filename;
-                mpack_write_cstr(writer, name != NULL ? PyUnicode_AsUTF8(name) : "NULL");
-                mpack_write_cstr(writer, filename != NULL ? PyUnicode_AsUTF8(filename) : "NULL");
-                mpack_write_u32(writer, frame->f_lineno);
+                // TODO: Check if frame is none or similar
+#if PY_VERSION_HEX >= 0x030a00a0
+                // Python >= 3.10 should use the standard accessors
+                PyCodeObject* code = PyFrame_GetCode(frame);
+                int lineno = PyFrame_GetLineNumber(frame);
+#else
+                PyCodeObject* code = frame->f_code;
+                int lineno = frame->f_lineno;
+#endif
+
+#if PY_VERSION_HEX >= 0x030b00a0
+                mpack_write_cstr(writer, Fprofile_UnicodeToUtf8(code->co_qualname));
+#else
+                mpack_write_cstr(writer, Fprofile_UnicodeToUtf8(code->co_name));
+#endif
+                mpack_write_cstr(writer, Fprofile_UnicodeToUtf8(code->co_filename));
+                mpack_write_u32(writer, lineno);
+
+#if PY_VERSION_HEX >= 0x030a00a0
+                Py_DECREF(code);
+#endif
             }
             mpack_finish_array(writer);
             break;
@@ -489,8 +523,20 @@ static int Fprofile_FunctionTrace(PyObject* obj, PyFrameObject* frame, int what,
             mpack_finish_array(writer);
 
             {
-                PyObject* name = frame->f_code->co_name;
-                mpack_write_cstr(writer, name != NULL ? PyUnicode_AsUTF8(name) : "NULL");
+#if PY_VERSION_HEX >= 0x030a00a0
+                PyCodeObject* code = PyFrame_GetCode(frame);
+#else
+                PyCodeObject* code = frame->f_code;
+#endif
+#if PY_VERSION_HEX >= 0x030b00a0
+                mpack_write_cstr(writer, Fprofile_UnicodeToUtf8(code->co_qualname));
+#else
+                mpack_write_cstr(writer, Fprofile_UnicodeToUtf8(code->co_name));
+#endif
+
+#if PY_VERSION_HEX >= 0x030a00a0
+                Py_DECREF(code);
+#endif
             }
             mpack_finish_array(writer);
             break;
